class PgAnonymizeException(Exception):
    """Base exception for all pganonymize errors."""


class InvalidFieldProvider(PgAnonymizeException):
    """Raised if an unknown field provider was used in the schema."""


class InvalidProvider(PgAnonymizeException):
    """Raised if an unknown provider class was requested."""


class InvalidProviderArgument(PgAnonymizeException):
    """Raised if an argument is unknown or invalid for a provider."""


class BadDataFormat(PgAnonymizeException):
    """Raised if the anonymized data cannot be copied."""



import argparse
import logging
import time

import yaml

from pganonymizer.utils import anonymize_tables, get_connection, truncate_tables


def main():
    parser = argparse.ArgumentParser(description='Anonymize data of a PostgreSQL database')
    parser.add_argument('-v', '--verbose', action='count', help='Increase verbosity')
    parser.add_argument('--schema', help='A YAML file that contains the anonymization rules', required=True,
                        default='./schema.yml')
    parser.add_argument('--dbname', help='Name of the database')
    parser.add_argument('--user', help='Name of the database user')
    parser.add_argument('--password', default='', help='Password for the database user')
    parser.add_argument('--host', help='Database hostname', default='localhost')
    parser.add_argument('--port', help='Port of the database', default='5432')
    parser.add_argument('--dry-run', action='store_true', help='Dont commit changes made on the database',
                        default=False)
    args = parser.parse_args()
    loglevel = logging.WARNING
    if args.verbose:
        loglevel = logging.DEBUG
    logging.basicConfig(format='%(levelname)s: %(message)s', level=loglevel)

    schema = yaml.load(open(args.schema), Loader=yaml.FullLoader)
    connection = get_connection(args)

    start_time = time.time()
    truncate_tables(connection, schema.get('truncate', []))
    anonymize_tables(connection, schema.get('tables', []), verbose=args.verbose)

    if not args.dry_run:
        connection.commit()
    connection.close()

    end_time = time.time()
    logging.info('Anonymization took {:.2f}s.'.format(end_time - start_time))


if __name__ == '__main__':
    main()



# Database arguments used for the CLI
DATABASE_ARGS = ('dbname', 'user', 'password', 'host', 'port')

# Default name for the primary key column
DEFAULT_PRIMARY_KEY = 'id'

# Delimiter used to buffer and import database data.
COPY_DB_DELIMITER = '\x1f'



import csv
import logging
import re
from cStringIO import StringIO

import psycopg2
import psycopg2.extras
from progress.bar import IncrementalBar
from psycopg2.errors import BadCopyFileFormat, InvalidTextRepresentation

from pganonymizer.constants import COPY_DB_DELIMITER, DATABASE_ARGS, DEFAULT_PRIMARY_KEY
from pganonymizer.exceptions import BadDataFormat
from pganonymizer.providers import get_provider


def anonymize_tables(connection, definitions, verbose=False):
    """
    Anonymize a list of tables according to the schema definition.

    :param connection: A database connection instance.
    :param list definitions: A list of table definitions from the YAML schema.
    :param bool verbose: Display logging information and a progress bar.
    """
    for definition in definitions:
        table_name = definition.keys()[0]
        logging.info('Found table definition "%s"', table_name)
        table_definition = definition[table_name]
        columns = table_definition.get('fields', [])
        excludes = table_definition.get('excludes', [])
        column_dict = get_column_dict(columns)
        primary_key = table_definition.get('primary_key', DEFAULT_PRIMARY_KEY)
        total_count = get_table_count(connection, table_name)
        data, table_columns = build_data(connection, table_name, columns, excludes, total_count, verbose)
        import_data(connection, column_dict, table_name, table_columns, primary_key, data)


def build_data(connection, table, columns, excludes, total_count, verbose=False):
    """
    Select all data from a table and build

    :param connection: A database connection instance.
    :param str table: Name of the table to retrieve the data.
    :param list columns:
    :param list[dict] excludes: A list of exclude definitions.
    :param int total_count: The amount of rows for the current table
    :param bool verbose: Display logging information and a progress bar.
    :return: A tuple containing the data list and a complete list of all table columns.
    :rtype: (list, list)
    """
    if verbose:
        bar = IncrementalBar('Anonymizing', max=total_count)
    sql = "SELECT * FROM {table};".format(table=table)
    cursor = connection.cursor(cursor_factory=psycopg2.extras.DictCursor, name='fetch_large_result')
    cursor.execute(sql)
    data = []
    table_columns = None
    while True:
        records = cursor.fetchmany(size=2000)
        if not records:
            break
        for row in records:
            row_column_dict = {}
            if not row_matches_excludes(row, excludes):
                row_column_dict = get_column_values(row, columns)
                for key, value in row_column_dict.items():
                    row[key] = value
            if verbose:
                bar.next()
            table_columns = row.keys()
            if not row_column_dict:
                continue
            data.append(row.values())
    if verbose:
        bar.finish()
    cursor.close()
    return data, table_columns


def row_matches_excludes(row, excludes=None):
    """
    Check whether a row matches a list of field exclusion patterns.

    :param list row: The data row
    :param list excludes: A list of field exclusion roles, e.g.:

        [
            {'email': ['\\S.*@example.com', '\\S.*@foobar.com', ]}
        ]

    :return: True or False
    :rtype: bool
    """
    excludes = excludes if excludes else []
    for definition in excludes:
        column = definition.keys()[0]
        for exclude in definition.get(column, []):
            pattern = re.compile(exclude, re.IGNORECASE)
            if pattern.match(row[column]):
                return True
    return False


def copy_from(connection, data, table, columns):
    """
    Copy the data from a table to a temporary table.

    :param connection: A database connection instance.
    :param list data: The data of a table.
    :param str table: Name of the temporary table used for copying the data.
    :param list columns: All columns of the current table.
    :raises BadDataFormat: If the data cannot be imported due to a invalid format.
    """
    new_data = data2csv(data)
    cursor = connection.cursor()
    try:
        cursor.copy_from(new_data, table, sep=COPY_DB_DELIMITER, null='\\N', columns=columns)
    except (BadCopyFileFormat, InvalidTextRepresentation) as e:
        raise BadDataFormat(e)
    cursor.close()


def import_data(connection, column_dict, source_table, table_columns, primary_key, data):
    """
    Import the temporary and anonymized data to a temporary table and write the changes back.

    :param connection: A database connection instance.
    :param dict column_dict:
    :param str source_table: Name of the table to be anonymized.
    :param list table_columns: A list of all table columns.
    :param str primary_key: Name of the tables primary key.
    :param list data: The table data.
    """
    primary_key = primary_key if primary_key else DEFAULT_PRIMARY_KEY
    cursor = connection.cursor()
    cursor.execute('CREATE TEMP TABLE source(LIKE %s INCLUDING ALL) ON COMMIT DROP;' % source_table)
    copy_from(connection, data, 'source', table_columns)
    set_columns = ', '.join(['{column} = s.{column}'.format(column=key) for key in column_dict.keys()])
    sql = (
        'UPDATE {table} t '
        'SET {columns} '
        'FROM source s '
        'WHERE t.{primary_key} = s.{primary_key}'
    ).format(table=source_table, primary_key=primary_key, columns=set_columns)
    cursor.execute(sql)
    cursor.execute('DROP TABLE source;')
    cursor.close()


def get_connection(args):
    """
    Return a connection to the database.

    :param args:
    :return: Connection instance
    :rtype: psycopg2.connection
    """
    pg_args = ({name: value for name, value in
                zip(DATABASE_ARGS, (args.dbname, args.user, args.password, args.host, args.port))})
    return psycopg2.connect(**pg_args)


def get_table_count(connection, table):
    """
    Return the number of table entries.

    :param connection: A database connection instance
    :param str table: Name of the database table
    :return: The number of table entries
    :rtype: int
    """
    sql = 'SELECT COUNT(*) FROM {table};'.format(table=table)
    cursor = connection.cursor()
    cursor.execute(sql)
    total_count = cursor.fetchone()[0]
    cursor.close()
    return total_count


def data2csv(data):
    """
    Return a string buffer, that contains delimited data.

    :param list data: A list of values
    :return: A stream that contains tab delimited csv data
    :rtype: StringIO
    """
    buf = StringIO()
    writer = csv.writer(buf, delimiter=COPY_DB_DELIMITER, lineterminator='\n', quotechar='~')
    [writer.writerow([(x is None and '\\N' or x) for x in row]) for row in data]
    buf.seek(0)
    return buf


def get_column_dict(columns):
    """
    Return a dictionary with all fields from the table definition and None as value.

    :param list columns: A list of field definitions from the YAML schema, e.g.:

        [
            {'first_name': {'provider': 'set', 'value': 'Foo'}},
            {'guest_email': {'append': '@localhost', 'provider': 'md5'}},
        ]
    :return: A dictionary containing all fields to be altered with a default value of None, e.g.:

        {'guest_email': None}

    :rtype: dict
    """
    column_dict = {}
    for definition in columns:
        column_name = definition.keys()[0]
        column_dict[column_name] = None
    return column_dict


def get_column_values(row, columns):
    """
    Return a dictionary for a single data row, with altered data.

    :param psycopg2.extras.DictRow row: A data row from the current table to be altered
    :param list columns: A list of table columns with their provider rules, e.g.:

        [
            {'guest_email': {'append': '@localhost', 'provider': 'md5'}}
        ]

    :return: A dictionary with all fields that have to be altered and their value for a single data row, e.g.:

        {'guest_email': '12faf5a9bb6f6f067608dca3027c8fcb@localhost'}

    :rtype: dict
    """
    column_dict = {}
    for definition in columns:
        column_name = definition.keys()[0]
        column_definition = definition[column_name]
        provider_config = column_definition.get('provider')
        orig_value = row.get(column_name)
        if not orig_value:
            # Skip the current column if there is no value to be altered
            continue
        provider = get_provider(provider_config)
        value = provider.alter_value(orig_value)
        append = column_definition.get('append')
        if append:
            value = value + append
        column_dict[column_name] = value
    return column_dict


def truncate_tables(connection, tables):
    """
    Truncate a list of tables.

    :param connection: A database connection instance
    :param list[str] tables: A list of table names
    """
    cursor = connection.cursor()
    for table_name in tables:
        logging.info('Truncating table "%s"', table_name)
        cursor.execute('TRUNCATE TABLE {table}'.format(table=table_name))
    cursor.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import distutils
import subprocess
from os.path import dirname, join

from setuptools import setup, find_packages


def read(*args):
    return open(join(dirname(__file__), *args)).read()


class ToxTestCommand(distutils.cmd.Command):
    """Distutils command to run tests via tox with 'python setup.py test'.

    Please note that in our standard configuration tox uses the dependencies in
    `requirements/dev.txt`, the list of dependencies in `tests_require` in
    `setup.py` is ignored!

    See https://docs.python.org/3/distutils/apiref.html#creating-a-new-distutils-command
    for more documentation on custom distutils commands.
    """
    description = "Run tests via 'tox'."
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.announce("Running tests with 'tox'...", level=distutils.log.INFO)
        return subprocess.call(['tox'])


exec(read('pganonymizer', 'version.py'))

install_requires = [
    'faker',
    'six',
    'progress',
    'psycopg2',
    'pyyaml'
]

tests_require = [
    'coverage',
    'flake8',
    'pydocstyle',
    'pylint',
    'pytest-pep8',
    'pytest-cov',
    'pytest-pythonpath',
    'pytest',
]

setup(
    name='pganonymize',
    version=__version__,  # noqa
    description='Commandline tool to anonymize PostgreSQL databases',
    long_description=read('README.rst'),
    author='Henning Kage',
    author_email='henning.kage@rheinwerk-verlag.de',
    maintainer='Rheinwerk Verlag GmbH Webteam',
    maintainer_email='webteam@rheinwerk-verlag.de',
    url='https://github.com/rheinwerk-verlag/postgresql-anonymizer',
    license='Proprietary',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Environment :: Console',
        'Topic :: Database'
    ],
    packages=find_packages(include=['pganonymizer*']),
    include_package_data=True,
    install_requires=install_requires,
    tests_require=tests_require,
    cmdclass={
        'test': ToxTestCommand,
    },
    entry_points={
        'console_scripts': [
            'pganonymize=pganonymizer.cli:main'
        ]
    }
)



import random
from hashlib import md5

from faker import Faker
from six import with_metaclass

from pganonymizer.exceptions import InvalidProvider, InvalidProviderArgument


PROVIDERS = []

fake_data = Faker()


def get_provider(provider_config):
    """
    Return a provider instance, according to the schema definition of a field.

    :param dict provider_config: A provider configuration for a single field, e.g.:

        {'name': 'set', 'value': 'Foo'}

    :return: A provider instance
    :rtype: Provider
    """
    def get_provider_class(cid):
        for cls in PROVIDERS:
            if cls.matches(name):
                return cls
    name = provider_config['name']
    cls = get_provider_class(name)
    if cls is None:
        raise InvalidProvider('Could not find provider with id %s' % name)
    return cls(**provider_config)


class ProviderMeta(type):
    """Metaclass to register all provider classes."""

    def __new__(cls, clsname, bases, attrs):
        newclass = super(ProviderMeta, cls).__new__(cls, clsname, bases, attrs)
        if clsname != 'Provider':
            PROVIDERS.append(newclass)
        return newclass


class Provider(object):
    """Base class for all providers."""

    id = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @classmethod
    def matches(cls, name):
        return cls.id.lower() == name.lower()

    def alter_value(self, value):
        raise NotImplementedError


class ChoiceProvider(with_metaclass(ProviderMeta, Provider)):
    """Provider that returns a random value from a list of choices."""

    id = 'choice'

    def alter_value(self, value):
        return random.choice(self.kwargs.get('values'))


class ClearProvider(with_metaclass(ProviderMeta, Provider)):
    """Provider to set a field value to None."""

    id = 'clear'

    def alter_value(self, value):
        return None


class FakeProvider(with_metaclass(ProviderMeta, Provider)):
    """Provider to generate fake data."""

    id = 'fake'

    @classmethod
    def matches(cls, name):
        return cls.id.lower() == name.split('.')[0].lower()

    def alter_value(self, value):
        func_name = self.kwargs['name'].split('.')[1]
        try:
            func = getattr(fake_data, func_name)
        except AttributeError as exc:
            raise InvalidProviderArgument(exc)
        return func()


class MaskProvider(with_metaclass(ProviderMeta, Provider)):
    """Provider that masks the original value."""

    id = 'mask'
    default_sign = 'X'

    def alter_value(self, value):
        sign = self.kwargs.get('sign', self.default_sign) or self.default_sign
        return sign * len(value)


class MD5Provider(with_metaclass(ProviderMeta, Provider)):
    """Provider to hash a value with the md5 algorithm."""

    id = 'md5'

    def alter_value(self, value):
        return md5(value.encode('utf-8')).hexdigest()


class SetProvider(with_metaclass(ProviderMeta, Provider)):
    """Provider to set a static value."""

    id = 'set'

    def alter_value(self, value):
        return self.kwargs.get('value')
